#include "TestCacheHeader.mqh"
#include "TstHeader.mqh"
#include "ExpTradeSummary.mqh"
#include "TestCacheInput.mqh"
#include "Mathematics.mqh"
#include "TestCacheRecord.mqh"
#include "TestCacheSymbolRecord.mqh"

template <typename T,typename U>
class TESTERCACHE_BASE
{
private:
  bool IsCorrectType( void ) const
  {
    return((typename(T) == "MATHEMATICS") && (this.Header.symbol[] == "")) ||
           ((typename(T) == "TestCacheSymbolRecord") && (this.Header.opt_mode == 2)) ||
           ((typename(T) == "ExpTradeSummary") && (this.Header.symbol[] != "") && (this.Header.opt_mode < 2));
  }
  
  void LoadHeader(const int handle)
  {
    ::FileReadStruct(handle, this.Header);
  }
  
  void LoadHeader(const uchar &Bytes[],int &Pos)
  {
    _W(this.Header) = Bytes;
    Pos += sizeof(this.Header);
  }

  virtual void LoadInputs(const int handle) = NULL;
  virtual void LoadInputs(const uchar &Bytes[],int &Pos) = NULL;

  virtual void LoadParameters(const int handle) = NULL;
  virtual void LoadParameters(const uchar &Bytes[],int &Pos) = NULL;

  virtual void LoadPasses(const int handle) = NULL;
  virtual void LoadPasses(const uchar &Bytes[],int &Pos) = NULL;

public:
  U Header;
  TestCacheInput Inputs[];
  uchar ParametersBuffer[];
  TestCacheRecord<T> Record[];
  int UnknownNum;

  bool Load( const string FileName )
  {
    const int handle = ::FileOpen(FileName, FILE_READ | FILE_BIN);
    bool Res = (handle != INVALID_HANDLE);

    if (Res)
    {
      ::ArrayFree(this.Inputs);
      ::ArrayFree(this.ParametersBuffer);
      ::ArrayFree(this.Record);
      this.UnknownNum = 0;

      LoadHeader(handle);

      if (Res = this.IsCorrectType())
      {
        LoadInputs(handle);
        LoadParameters(handle);
        LoadPasses(handle);
      }

      ::FileClose(handle);
    }

    return(Res);
  }

  bool Load( const uchar &Bytes[] )
  {
    bool Res = ::ArraySize(Bytes);

    if (Res)
    {
      ::ArrayFree(this.Inputs);
      ::ArrayFree(this.ParametersBuffer);
      ::ArrayFree(this.Record);
      this.UnknownNum = 0;

      int Pos=0;
      LoadHeader(Bytes, Pos);

      if (Res = this.IsCorrectType())
      {
        LoadInputs(Bytes, Pos);
        LoadParameters(Bytes, Pos);
        LoadPasses(Bytes, Pos);
      }
    }

    return(Res);
  }

  bool Save( const string FileName )
  {
    const int handle = ::FileOpen(FileName, FILE_WRITE | FILE_BIN);
    bool Res = (handle != INVALID_HANDLE);

    if (Res)
    {
      ::FileWriteStruct(handle, this.Header);

      ::FileWriteArray(handle, this.Inputs);

      ::FileWriteArray(handle, this.ParametersBuffer);

      if (this.Header.opt_mode == 0)
      {
        if (this.Header.snapshot_size)
        {
          uchar Buffer[];

          ::ArrayResize(Buffer, sizeof(int) * (this.Header.snapshot_size - 1));
          ::ArrayInitialize(Buffer, (char)0xFF);

          ::FileWriteArray(handle, Buffer);
        }

        ::FileWriteInteger(handle, this.UnknownNum);
      }

      const int Size = ::ArraySize(this.Record);

      for (int i = 0; i < Size; i++)
      {
        ::FileWriteStruct(handle, this.Record[i].Result);

        ::FileWriteArray(handle, this.Record[i].OptBuffer);

//        if (!this.Header.passes_total)
          ::FileWriteArray(handle, this.Record[i].Genetic);
      }

      ::FileClose(handle);
    }

    return(Res);
  }

  TESTERCACHE_BASE( void ) : UnknownNum(0)
  {
  }

  TESTERCACHE_BASE( const string FileName ) : UnknownNum(0)
  {
    this.Load(FileName);
  }

  // https://www.mql5.com/ru/forum/324536/page20#comment_13850494
  TESTERCACHE_BASE( const uchar &Bytes[] ) : UnknownNum(0)
  {
    this.Load(Bytes);
  }

  const T operator[]( const int Num ) const
  {
    return(this.Record[Num].Result);
  }

  int GetAmount( void ) const
  {
    return(::ArraySize(this.Record));
  }
};

template <typename T>
class TESTERCACHE : public TESTERCACHE_BASE<T,TestCacheHeader>
{
private:
  virtual void LoadInputs(const int handle)
  {
    ::FileReadArray(handle, this.Inputs, 0, this.Header.parameters_total);
  }
  
  virtual void LoadInputs(const uchar &Bytes[],int &Pos)
  {
    Pos += ::_ArrayCopy(this.Inputs, Bytes, 0, Pos, this.Header.parameters_total * sizeof(TestCacheInput));
  }

  virtual void LoadParameters(const int handle)
  {
    ::FileReadArray(handle, this.ParametersBuffer, 0, this.Header.parameters_size);
  }
  
  virtual void LoadParameters(const uchar &Bytes[],int &Pos)
  {
    Pos += ::_ArrayCopy(this.ParametersBuffer, Bytes, 0, Pos, this.Header.parameters_size);
  }

  virtual void LoadPasses(const int handle)
  {
    const int Size = ::ArrayResize(this.Record, this.Header.passes_passed);

    if (this.Header.opt_mode == 0)
      this.UnknownNum = ::FileReadInteger(handle);

    for (int i = 0; i < Size; i++)
    {
      ::FileReadStruct(handle, this.Record[i].Result);

      ::FileReadArray(handle, this.Record[i].OptBuffer, 0, this.Header.opt_params_size);

      if (this.Header.opt_mode == 1)
        ::FileReadArray(handle, this.Record[i].Genetic, 0, sizeof(long) + this.Header.dwords_cnt * sizeof(int));
      else
        ::FileReadArray(handle, this.Record[i].Genetic, 0, this.Header.dwords_cnt * sizeof(int));
    }
  }
  
  virtual void LoadPasses(const uchar &Bytes[],int &Pos)
  {
    const int Size = ::ArrayResize(this.Record, this.Header.passes_passed);

    if (this.Header.opt_mode == 0)
    {
      this.UnknownNum = _R(Bytes)[Pos];

      Pos += sizeof(this.UnknownNum);
    }

    for (int i = 0; i < Size; i++)
    {
      T Result[1];

      Pos += ::_ArrayCopy(Result, Bytes, 0, Pos, sizeof(T));
      this.Record[i].Result = Result[0];

      Pos += ::_ArrayCopy(this.Record[i].OptBuffer, Bytes, 0, Pos, this.Header.opt_params_size);

      if (this.Header.opt_mode == 1)
        Pos += ::_ArrayCopy(this.Record[i].Genetic, Bytes, 0, Pos, sizeof(long) + this.Header.dwords_cnt * sizeof(int));
      else
        Pos += ::_ArrayCopy(this.Record[i].Genetic, Bytes, 0, Pos, this.Header.dwords_cnt * sizeof(int));
    }
  }

public:
#define PARAM_NAME 0
#define PARAM_VALUE 1
#define PARAM_START 2
#define PARAM_STEP 3
#define PARAM_STOP 4

#define MACROS(A, B)                                                                                                      \
  const string Str##A = ((Type == TYPE_FLOAT) || (Type == TYPE_DOUBLE)) ? (string)this.Inputs[i].StartStepStop.Number.##A \
                                                                      : (string)this.Inputs[i].StartStepStop.Integer.##A; \
  Params[j][PARAM_##B].type = Params[j][PARAM_VALUE].type;                                                                \
  Params[j][PARAM_##B].integer_value = (long)Str##A;                                                                      \
  Params[j][PARAM_##B].double_value = (double)Str##A;                                                                     \
  Params[j][PARAM_##B].string_value = Str##A;

  int GetInputs( const int Num, MqlParam &Params[][5], const bool Full = false ) const
  {
    int j = 0;
    int Count = 0;
    const int Size = ::ArraySize(this.Inputs);

    ::ArrayResize(Params, Size);
    ::ZeroMemory(Params);

    for (int i = 0; i < Size; i++)
      if (Full || this.Inputs[i].flag)
      {
        Params[j][PARAM_NAME].type = TYPE_STRING;
        Params[j][PARAM_NAME].string_value = this.Inputs[i].name[];

        const int Type = this.Inputs[i].type - TYPE_OFFSET;
        string StrValue;

        if (this.Inputs[i].flag)
        {
          StrValue = ((Type == TYPE_FLOAT) || (Type == TYPE_DOUBLE)) ? (string)_R(this.Record[Num].OptBuffer)[(double)Count]
                                                                     : (string)_R(this.Record[Num].OptBuffer)[(long)Count];

          Count += sizeof(long);
        }
        else
        {
          const string Str = this.Inputs[i].ToString(this.ParametersBuffer);

          StrValue = (Type == TYPE_BOOL) ? ((Str == "true") ? "1" : "0") :
                     ((Type == TYPE_COLOR) ? (string)(long)(color)Str :
                     ((Type == TYPE_DATETIME) ? (string)(long)(datetime)Str : Str));
        }

        Params[j][PARAM_VALUE].type = (Type >= TYPE_BOOL) && (Type <= TYPE_STRING) ? (ENUM_DATATYPE)Type : TYPE_INT;
        Params[j][PARAM_VALUE].integer_value = (long)StrValue;
        Params[j][PARAM_VALUE].double_value = (double)StrValue;
        Params[j][PARAM_VALUE].string_value = StrValue;

        MACROS(Start, START)
        MACROS(Step, STEP)
        MACROS(Stop, STOP)

        j++;
      }

    return(::ArrayResize(Params, j));
  }
#undef MACROS

  bool SaveSet( const int Num, string FileName = NULL, const bool Details = true )
  {
    if (FileName == NULL)
      FileName = this.Header.expert_name[] + "_" + (string)Num + ".set";

    const int handle = ::FileOpen(FileName, FILE_WRITE | FILE_UNICODE | FILE_TXT);
    const bool Res= (handle != INVALID_HANDLE);

    if (Res)
    {
      MqlParam Params[][5];
      const int Size = this.GetInputs(Num, Params, true);

      if (Details)
      {
        ::FileWriteString(handle, "; saved on " + (string)::TimeLocal() + "\n; " +
                                  this.Header.expert_path[] +"\n; " +
                                  this.Header.symbol[] +"\n; " +
                                  ::TimeToString(this.Header.date_from, TIME_DATE) + " - " + ::TimeToString(this.Header.date_to, TIME_DATE) + "\n; " +
                                  ::DoubleToString(this[Num].TesterStatistics(STAT_PROFIT), 0) + ", " +
                                  ::DoubleToString(this[Num].TesterStatistics(STAT_TRADES), 0) + ", " +
                                  ::DoubleToString(this[Num].TesterStatistics(STAT_PROFIT_FACTOR), 2) + ", " +
                                  ::DoubleToString(this[Num].TesterStatistics(STAT_EXPECTED_PAYOFF), 2) +  ", -" +
                                  ::DoubleToString(this[Num].TesterStatistics(STAT_EQUITY_DD), 2) +
                                  "\n;\n");

      }

      for (int i = 0; i < Size; i++)
        ::FileWriteString(handle, Params[i][PARAM_NAME].string_value == "" ? ";\n" :
                                  Params[i][PARAM_NAME].string_value + "=" +
                                  Params[i][PARAM_VALUE].string_value +
                                  ((!Params[i][PARAM_START].double_value &&
                                    !Params[i][PARAM_STEP].double_value &&
                                    !Params[i][PARAM_STOP].double_value) ? NULL : "||" +
                                    Params[i][PARAM_START].string_value + "||" +
                                    Params[i][PARAM_STEP].string_value + "||" +
                                    Params[i][PARAM_STOP].string_value + "||" +
                                    (this.Inputs[i].flag ? "Y" : "N")) + "\n");

      if (Details)
      {
        string Str = ";\n" + this[Num].ToString() + "\n" + this.Header.ToString();

        ::StringReplace(Str, "\n", "\n; ");

        ::FileWriteString(handle, Str);
      }

      ::FileClose(handle);
    }

    return(Res);
  }
};

template <typename T>
class SINGLEPASSTST : public TESTERCACHE_BASE<T,TstHeader>
{
private:
  virtual void LoadInputs(const int handle)
  {
    string InputsString=::FileReadString(handle, this.Header.parameters_size);
    DebugBreak();
  }
  
  virtual void LoadInputs(const uchar &Bytes[],int &Pos)
  {
    short InputsArray[];
    Pos += ::_ArrayCopy(InputsArray, Bytes, 0, Pos, this.Header.parameters_size*sizeof(InputsArray[0]));
    string InputsString=::ShortArrayToString(InputsArray);
  }

  virtual void LoadParameters(const int handle)
  {
    uchar TempBuffer[104];
    ::FileReadArray(handle, TempBuffer, 0, sizeof(TempBuffer));
  }
  
  virtual void LoadParameters(const uchar &Bytes[],int &Pos)
  {
    uchar TempBuffer[104];
    Pos += ::_ArrayCopy(TempBuffer, Bytes, 0, Pos, sizeof(TempBuffer));
  }

  virtual void LoadPasses(const int handle)
  {
    ::ArrayResize(this.Record, 1);
    ::FileReadStruct(handle, this.Record[0].Result);
  }
  
  virtual void LoadPasses(const uchar &Bytes[],int &Pos)
  {
    ::ArrayResize(this.Record, 1);

    T Result[1];
    Pos += ::_ArrayCopy(Result, Bytes, 0, Pos, sizeof(T));
    this.Record[0].Result = Result[0];
  }
};